"""OCI Node Provider.

Node provider is called by the Ray Autoscaler to provision new compute
resources (head / worker nodes).

To show debug messages, export SKYPILOT_DEBUG=1 

History:
 - Hysun He (hysun.he@oracle.com) @ Apr, 2023: Initial implementation
 
"""

import copy
from datetime import datetime
import logging
import threading
import time

from ray.autoscaler.node_provider import NodeProvider
from ray.autoscaler.tags import TAG_RAY_CLUSTER_NAME
from ray.autoscaler.tags import TAG_RAY_LAUNCH_CONFIG
from ray.autoscaler.tags import TAG_RAY_NODE_KIND
from ray.autoscaler.tags import TAG_RAY_USER_NODE_TYPE

from sky.adaptors import oci as oci_adaptor
from sky.clouds.utils import oci_utils
from sky.skylet.providers.oci import utils
from sky.skylet.providers.oci.query_helper import oci_query_helper

logger = logging.getLogger(__name__)


def synchronized(f):

    def wrapper(self, *args, **kwargs):
        self.lock.acquire()
        try:
            return f(self, *args, **kwargs)
        finally:
            self.lock.release()

    return wrapper


class OCINodeProvider(NodeProvider):
    """Node Provider for OracleCloud (OCI)."""

    def __init__(self, provider_config, cluster_name):
        NodeProvider.__init__(self, provider_config, cluster_name)
        self.lock = threading.RLock()
        self.cached_nodes = {}
        self.cache_stopped_nodes = provider_config.get("cache_stopped_nodes",
                                                       True)
        self.region = provider_config["region"]

        # Do a read-ahead cache loading to improve performance.
        self._get_filtered_nodes({})

    @synchronized
    def _get_filtered_nodes(self, tag_filters, force=False):
        # Make sure the cluster_name is always an criterion
        tag_filters = {**tag_filters, TAG_RAY_CLUSTER_NAME: self.cluster_name}

        return_nodes = {}
        if not force:
            # Query cache first to reduce API call.
            cache_hit = False
            for k, node in self.cached_nodes.items():
                tags = node["tags"]
                unmatched_tags = [
                    k for k, v in tag_filters.items()
                    if k not in tags or v != tags[k]
                ]
                if len(unmatched_tags) == 0:
                    return_nodes[k] = node
                    cache_hit |= True

            if cache_hit:
                return return_nodes

        insts = oci_query_helper.query_instances_by_tags(
            tag_filters, self.region)
        for inst in insts:
            inst_id = inst.identifier
            if inst_id in self.cached_nodes:
                del self.cached_nodes[inst_id]

            item = self.get_inst_obj({
                "inst_id": inst_id,
                "ad": inst.availability_domain,
                "compartment": inst.compartment_id,
                "lifecycle_state": inst.lifecycle_state,
                "oci_tags": inst.freeform_tags,
            })
            return_nodes[inst_id] = item
            self.cached_nodes[inst_id] = item

        return return_nodes

    @utils.debug_enabled(logger=logger)
    def non_terminated_nodes(self, tag_filters):
        """Return a list of node ids filtered by the specified tags dict.

        This list must not include terminated nodes. For performance reasons,
        providers are allowed to cache the result of a call to
        non_terminated_nodes() to serve single-node queries
        (e.g. is_running(node_id)). This means that non_terminated_nodes()
        must be called again to refresh results.
        """
        VALIDITY_TAGS = [
            TAG_RAY_CLUSTER_NAME,
            TAG_RAY_NODE_KIND,
            TAG_RAY_USER_NODE_TYPE,
            TAG_RAY_LAUNCH_CONFIG,
        ]
        filters = {
            tag: tag_filters[tag] for tag in VALIDITY_TAGS if tag in tag_filters
        }

        nodes = self._get_filtered_nodes(tag_filters=filters)
        return [k for k, v in nodes.items() if v["status"] == "RUNNING"]

    @utils.debug_enabled(logger=logger)
    def is_running(self, node_id):
        """Return whether the specified node is running."""
        node = self._get_cached_node(node_id=node_id)
        check_result = node is None or node["status"] == "RUNNING"

        return check_result

    @utils.debug_enabled(logger=logger)
    def is_terminated(self, node_id):
        """Return whether the specified node is terminated."""
        node = self._get_cached_node(node_id=node_id)
        check_result = ((node is None) or (node["status"] == "TERMINATED") or
                        (node["status"] == "TERMINATING"))

        return check_result

    @utils.debug_enabled(logger=logger)
    def node_tags(self, node_id):
        return self.cached_nodes[node_id]["tags"]

    @utils.debug_enabled(logger=logger)
    def external_ip(self, node_id):
        """Returns the external ip of the given node."""
        return self._get_cached_node(node_id=node_id)["external_ip"]

    @utils.debug_enabled(logger=logger)
    def internal_ip(self, node_id):
        """Returns the internal ip (Ray ip) of the given node."""
        return self._get_cached_node(node_id=node_id)["internal_ip"]

    @synchronized
    @utils.debug_enabled(logger=logger)
    def create_node(self, node_config, tags, count):
        """Creates a number of nodes within the namespace."""
        start_time = round(time.time() * 1000)
        starting_insts = []
        # Check first if it neccessary to create new nodes / start stopped nodes
        VALIDITY_TAGS = [
            TAG_RAY_CLUSTER_NAME,
            TAG_RAY_NODE_KIND,
            TAG_RAY_USER_NODE_TYPE,
        ]
        filters = {tag: tags[tag] for tag in VALIDITY_TAGS if tag in tags}

        # Starting stopped nodes if cache_stopped_nodes=True
        if self.cache_stopped_nodes:
            logger.debug("Checking existing stopped nodes.")

            filters_with_launch_config = copy.copy(filters)
            if TAG_RAY_LAUNCH_CONFIG in tags:
                filters_with_launch_config[TAG_RAY_LAUNCH_CONFIG] = tags[
                    TAG_RAY_LAUNCH_CONFIG]

            nodes_matching_launch_config = self.stopped_nodes(
                filters_with_launch_config)
            logger.debug(f"Found stopped nodes (with same launch config): "
                         f"{len(nodes_matching_launch_config)}")

            reuse_nodes = []
            if len(nodes_matching_launch_config) >= count:
                reuse_nodes = nodes_matching_launch_config[:count]
            else:
                nodes_all = self.stopped_nodes(filters)
                logger.debug(f"Found stopped nodes (regardless launch config): "
                             f"{len(nodes_all)}")
                nodes_matching_launch_config_ids = [
                    n["id"] for n in nodes_matching_launch_config
                ]
                nodes_non_matching_launch_config = [
                    n for n in nodes_all
                    if n["id"] not in nodes_matching_launch_config_ids
                ]
                reuse_nodes = (nodes_matching_launch_config +
                               nodes_non_matching_launch_config)
                reuse_nodes = reuse_nodes[:count]

            logger.info(
                f"Reusing nodes {len(reuse_nodes)}: {list(reuse_nodes)}. "
                "To disable reuse, set `cache_stopped_nodes: False` "
                "under `provider` in the cluster configuration.",)

            for reuse_node in reuse_nodes:
                if reuse_node["status"] == "STOPPING":
                    get_instance_response = oci_adaptor.get_core_client(
                        self.region,
                        oci_utils.oci_config.get_profile()).get_instance(
                            instance_id=reuse_node["id"])
                    oci_adaptor.get_oci().wait_until(
                        oci_adaptor.get_core_client(
                            self.region, oci_utils.oci_config.get_profile()),
                        get_instance_response,
                        "lifecycle_state",
                        "STOPPED",
                    )

            start_time1 = round(time.time() * 1000)
            for matched_node in reuse_nodes:
                matched_node_id = matched_node["id"]
                instance_action_response = oci_adaptor.get_core_client(
                    self.region,
                    oci_utils.oci_config.get_profile()).instance_action(
                        instance_id=matched_node_id, action="START")

                starting_inst = instance_action_response.data
                starting_insts.append({
                    "inst_id": starting_inst.id,
                    "ad": starting_inst.availability_domain,
                    "compartment": starting_inst.compartment_id,
                    "lifecycle_state": starting_inst.lifecycle_state,
                    "oci_tags": starting_inst.freeform_tags,
                })
            count -= len(reuse_nodes)

            launch_stopped_time = round(time.time() * 1000) - start_time1
            logger.debug(
                "Time elapsed(Launch stopped): {0} milli-seconds.".format(
                    launch_stopped_time))
        # end if self.cache_stopped_nodes:...

        # Let's create additional new nodes (if neccessary)
        if count > 0:
            compartment = oci_query_helper.find_compartment(self.region)
            vcn = oci_query_helper.find_create_vcn_subnet(self.region)
            if vcn is None:
                raise RuntimeError("VcnSubnetNotFound Error!")

            ocpu_count = 0
            vcpu_str = node_config["VCPUs"]
            instance_type_str = node_config["InstanceType"]

            if vcpu_str is not None and vcpu_str != "None":
                if instance_type_str.startswith(
                        f"{oci_utils.oci_config.VM_PREFIX}.A"):
                    # For ARM cpu, 1*ocpu = 1*vcpu
                    ocpu_count = round(float(vcpu_str))
                else:
                    # For Intel / AMD cpu, 1*ocpu = 2*vcpu
                    ocpu_count = round(float(vcpu_str) / 2)
            ocpu_count = 1 if (ocpu_count > 0 and
                               ocpu_count < 1) else ocpu_count

            machine_shape_config = None
            if ocpu_count > 0:
                mem = node_config["MemoryInGbs"]
                if mem is not None and mem != "None":
                    machine_shape_config = (oci_adaptor.get_oci().core.models.
                                            LaunchInstanceShapeConfigDetails(
                                                ocpus=ocpu_count,
                                                memory_in_gbs=mem))
                else:
                    machine_shape_config = (oci_adaptor.get_oci().core.models.
                                            LaunchInstanceShapeConfigDetails(
                                                ocpus=ocpu_count))

            preempitible_config = (oci_adaptor.get_oci(
            ).core.models.PreemptibleInstanceConfigDetails(
                preemption_action=oci_adaptor.get_oci().core.models.
                TerminatePreemptionAction(type="TERMINATE",
                                          preserve_boot_volume=False))
                                   if node_config["Preemptible"] else None)

            logger.debug(f"Shape: {instance_type_str}, ocpu: {ocpu_count}")
            logger.debug(f"Shape config is {machine_shape_config}")
            logger.debug(f"Spot config is {preempitible_config}")

            vm_tags = {
                **tags,
                TAG_RAY_CLUSTER_NAME: self.cluster_name,
                "sky_spot_flag": str(node_config["Preemptible"]).lower(),
            }
            # Use UTC time so that header & worker nodes use same rule
            batch_id = datetime.utcnow().strftime("%Y%m%d%H%M%S")
            node_type = tags[TAG_RAY_NODE_KIND]

            oci_query_helper.subscribe_image(
                compartment_id=compartment,
                listing_id=node_config["AppCatalogListingId"],
                resource_version=node_config["ResourceVersion"],
                region=self.region,
            )

            start_time1 = round(time.time() * 1000)
            for seq in range(1, count + 1):
                launch_instance_response = oci_adaptor.get_core_client(
                    self.region, oci_utils.oci_config.get_profile()
                ).launch_instance(launch_instance_details=oci_adaptor.get_oci(
                ).core.models.LaunchInstanceDetails(
                    availability_domain=node_config["AvailabilityDomain"],
                    compartment_id=compartment,
                    shape=instance_type_str,
                    display_name=
                    f"{self.cluster_name}_{node_type}_{batch_id}_{seq}",
                    freeform_tags=vm_tags,
                    metadata={
                        "ssh_authorized_keys": node_config["AuthorizedKey"]
                    },
                    source_details=oci_adaptor.get_oci(
                    ).core.models.InstanceSourceViaImageDetails(
                        source_type="image",
                        image_id=node_config["ImageId"],
                        boot_volume_size_in_gbs=node_config["BootVolumeSize"],
                        boot_volume_vpus_per_gb=int(
                            node_config["BootVolumePerf"]),
                    ),
                    create_vnic_details=oci_adaptor.get_oci(
                    ).core.models.CreateVnicDetails(
                        assign_public_ip=True,
                        subnet_id=vcn,
                    ),
                    shape_config=machine_shape_config,
                    preemptible_instance_config=preempitible_config,
                ))

                new_inst = launch_instance_response.data
                starting_insts.append({
                    "inst_id": new_inst.id,
                    "ad": new_inst.availability_domain,
                    "compartment": new_inst.compartment_id,
                    "lifecycle_state": new_inst.lifecycle_state,
                    "oci_tags": new_inst.freeform_tags,
                })
            # end for loop

            launch_new_time = round(time.time() * 1000) - start_time1
            logger.debug("Time elapsed(Launch): {0} milli-seconds.".format(
                launch_new_time))
        # end if count > 0:...

        for ninst in starting_insts:
            # Waiting for the instance to be RUNNING state
            get_instance_response = oci_adaptor.get_core_client(
                self.region, oci_utils.oci_config.get_profile()).get_instance(
                    instance_id=ninst["inst_id"])
            oci_adaptor.get_oci().wait_until(
                oci_adaptor.get_core_client(self.region,
                                            oci_utils.oci_config.get_profile()),
                get_instance_response,
                "lifecycle_state",
                "RUNNING",
            )
            ninst["lifecycle_state"] = "RUNNING"
            self.cached_nodes[ninst["inst_id"]] = self.get_inst_obj(ninst)

        total_time = round(time.time() * 1000) - start_time
        logger.debug(
            "Total time elapsed: {0} milli-seconds.".format(total_time))

    def get_inst_obj(self, inst_info):
        list_vnic_attachments_response = oci_adaptor.get_core_client(
            self.region,
            oci_utils.oci_config.get_profile()).list_vnic_attachments(
                availability_domain=inst_info["ad"],
                compartment_id=inst_info["compartment"],
                instance_id=inst_info["inst_id"],
            )

        vnic = list_vnic_attachments_response.data[0]
        get_vnic_response = (oci_adaptor.get_net_client(
            self.region, oci_utils.oci_config.get_profile()).get_vnic(
                vnic_id=vnic.vnic_id).data)

        internal_ip = get_vnic_response.private_ip
        external_ip = get_vnic_response.public_ip
        if external_ip is None:
            external_ip = internal_ip

        return {
            "id": inst_info["inst_id"],
            "external_ip": external_ip,
            "internal_ip": internal_ip,
            "tags": inst_info["oci_tags"],
            "status": inst_info["lifecycle_state"],
        }

    @synchronized
    @utils.debug_enabled(logger=logger)
    def set_node_tags(self, node_id, tags):
        existing_tags = self._get_cached_node(node_id)["tags"]
        combined_tags = dict(existing_tags, **tags)

        self.cached_nodes[node_id]["tags"] = combined_tags
        retry_count = 0
        while retry_count < oci_utils.oci_config.MAX_RETRY_COUNT:
            try:
                oci_adaptor.get_core_client(
                    self.region,
                    oci_utils.oci_config.get_profile()).update_instance(
                        instance_id=node_id,
                        update_instance_details=oci_adaptor.get_oci().core.
                        models.UpdateInstanceDetails(
                            freeform_tags=combined_tags),
                    )
                logger.info(f"Tags are well set for node {node_id}")
                break
            except Exception as e:
                retry_count = retry_count + 1
                wait_seconds = oci_utils.oci_config.RETRY_INTERVAL_BASE_SECONDS * retry_count
                logger.warn(
                    f"Not ready yet, wait {wait_seconds} seconds & retry!")
                logger.warn(f"Exception message is {str(e)}")
                time.sleep(wait_seconds)

    @synchronized
    def terminate_node(self, node_id):
        """Terminates the specified node."""
        logger.info(f"terminate_node {node_id}...")
        node = self._get_cached_node(node_id)
        if node is None:
            logger.info(f"The node is not existed: {node_id}..")
            return  # Node not exists yet.

        logger.debug(f"sky_spot_flag: {node['tags']['sky_spot_flag']}")
        preemptibleFlag = (True if node and
                           (str(node["tags"]["sky_spot_flag"]) == "true") else
                           False)

        if self.cache_stopped_nodes and not preemptibleFlag:
            logger.info(f"Stopping instance {node_id}"
                        "(to fully terminate instead, "
                        "set `cache_stopped_nodes: False` "
                        "under `provider` in the cluster configuration)")
            instance_action_response = oci_adaptor.get_core_client(
                self.region,
                oci_utils.oci_config.get_profile()).instance_action(
                    instance_id=node_id, action="STOP")
            logger.info(
                f"Stopped the instance {instance_action_response.data.id}")
            if node_id in self.cached_nodes:
                self.cached_nodes[node_id]["status"] = "STOPPED"
            state_word = "Stopped"
        else:
            terminate_instance_response = oci_adaptor.get_core_client(
                self.region,
                oci_utils.oci_config.get_profile()).terminate_instance(node_id)
            logger.debug(terminate_instance_response.data)
            if node_id in self.cached_nodes:
                del self.cached_nodes[node_id]
            state_word = "Terminated"

        logger.info(
            f"{state_word} {node_id} w/ sky_spot_flag: {preemptibleFlag}.")

    def _get_node(self, node_id):
        self._get_filtered_nodes({},
                                 force=True)  # All except for those terminated.
        return self.cached_nodes.get(node_id, None)

    def _get_cached_node(self, node_id):
        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]
        return self._get_node(node_id=node_id)

    def stopped_nodes(self, tag_filters):
        """Return a list of stopped nodes filtered by the specified tags dict."""
        nodes = self._get_filtered_nodes(tag_filters=tag_filters, force=True)
        return [
            v for _, v in nodes.items()
            if v["status"] in ("STOPPED", "STOPPING")
        ]

    def running_nodes(self, tag_filters):
        """Return a list of running node ids filtered by the specified tags dict."""
        nodes = self._get_filtered_nodes(tag_filters=tag_filters)
        return [k for k, v in nodes.items() if v["status"] == "RUNNING"]



"""Cudo Compute cloud adaptor."""

import functools

_cudo_sdk = None


def import_package(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _cudo_sdk
        if _cudo_sdk is None:
            try:
                import cudo_compute as _cudo  # pylint: disable=import-outside-toplevel
                _cudo_sdk = _cudo
            except ImportError:
                raise ImportError(
                    'Fail to import dependencies for Cudo Compute.'
                    'Try pip install "skypilot[cudo]"') from None
        return func(*args, **kwargs)

    return wrapper


@import_package
def cudo():
    """Return the Cudo Compute package."""
    return _cudo_sdk



"""Kubernetes enums for SkyPilot."""
import enum


class KubernetesNetworkingMode(enum.Enum):
    """Enum for the different types of networking modes for accessing
    jump pods.
    """
    NODEPORT = 'nodeport'
    PORTFORWARD = 'portforward'

    @classmethod
    def from_str(cls, mode: str) -> 'KubernetesNetworkingMode':
        """Returns the enum value for the given string."""
        if mode.lower() == cls.NODEPORT.value:
            return cls.NODEPORT
        elif mode.lower() == cls.PORTFORWARD.value:
            return cls.PORTFORWARD
        else:
            raise ValueError(f'Unsupported kubernetes networking mode: '
                             f'{mode}. The mode must be either '
                             f'\'{cls.PORTFORWARD.value}\' or '
                             f'\'{cls.NODEPORT.value}\'. ')


class KubernetesServiceType(enum.Enum):
    """Enum for the different types of services."""
    NODEPORT = 'NodePort'
    CLUSTERIP = 'ClusterIP'


class KubernetesPortMode(enum.Enum):
    """Enum for the different types of modes supported for opening
    ports on Kubernetes.
    """
    INGRESS = 'ingress'
    LOADBALANCER = 'loadbalancer'



"""Restarts skylet if version does not match"""

import os
import subprocess

from sky.skylet import constants

VERSION_FILE = os.path.expanduser(constants.SKYLET_VERSION_FILE)


def restart_skylet():
    # Kills old skylet if it is running.
    # TODO(zhwu): make the killing graceful, e.g., use a signal to tell
    # skylet to exit, instead of directly killing it.
    subprocess.run(
        'ps aux | grep "sky.skylet.skylet" | grep "python3 -m"'
        '| awk \'{print $2}\' | xargs kill >> ~/.sky/skylet.log 2>&1',
        shell=True,
        check=False)
    subprocess.run(
        'nohup python3 -m sky.skylet.skylet'
        ' >> ~/.sky/skylet.log 2>&1 &',
        shell=True,
        check=True)
    with open(VERSION_FILE, 'w', encoding='utf-8') as v_f:
        v_f.write(constants.SKYLET_VERSION)


proc = subprocess.run(
    'ps aux | grep -v "grep" | grep "sky.skylet.skylet" | grep "python3 -m"',
    shell=True,
    check=False)

running = (proc.returncode == 0)

version_match = False
found_version = None
if os.path.exists(VERSION_FILE):
    with open(VERSION_FILE, 'r', encoding='utf-8') as f:
        found_version = f.read().strip()
        if found_version == constants.SKYLET_VERSION:
            version_match = True

version_string = (f' (found version {found_version}, new version '
                  f'{constants.SKYLET_VERSION})')
if not running:
    print('Skylet is not running. Starting (version '
          f'{constants.SKYLET_VERSION})...')
elif not version_match:
    print(f'Skylet is stale{version_string}. Restarting...')
else:
    print(
        f'Skylet is running with the latest version {constants.SKYLET_VERSION}.'
    )

if not running or not version_match:
    restart_skylet()



"""Kubernetes."""
import json
import os
import typing
from typing import Dict, Iterator, List, Optional, Tuple

from sky import clouds
from sky import sky_logging
from sky.adaptors import kubernetes
from sky.clouds import service_catalog
from sky.provision.kubernetes import network_utils
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import common_utils
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    # Renaming to avoid shadowing variables.
    from sky import resources as resources_lib

logger = sky_logging.init_logger(__name__)

# Check if KUBECONFIG is set, and use it if it is.
DEFAULT_KUBECONFIG_PATH = '~/.kube/config'
CREDENTIAL_PATH = os.environ.get('KUBECONFIG', DEFAULT_KUBECONFIG_PATH)


@clouds.CLOUD_REGISTRY.register
class Kubernetes(clouds.Cloud):
    """Kubernetes."""

    SKY_SSH_KEY_SECRET_NAME = f'sky-ssh-{common_utils.get_user_hash()}'
    SKY_SSH_JUMP_NAME = f'sky-ssh-jump-{common_utils.get_user_hash()}'
    PORT_FORWARD_PROXY_CMD_TEMPLATE = \
        'kubernetes-port-forward-proxy-command.sh.j2'
    PORT_FORWARD_PROXY_CMD_PATH = '~/.sky/port-forward-proxy-cmd.sh'
    # Timeout for resource provisioning. This timeout determines how long to
    # wait for pod to be in pending status before giving up.
    # Larger timeout may be required for autoscaling clusters, since autoscaler
    # may take some time to provision new nodes.
    # Note that this timeout includes time taken by the Kubernetes scheduler
    # itself, which can be upto 2-3 seconds.
    # For non-autoscaling clusters, we conservatively set this to 10s.
    # TODO(romilb): Make the timeout configurable.
    TIMEOUT = 10

    _DEFAULT_NUM_VCPUS = 2
    _DEFAULT_MEMORY_CPU_RATIO = 1
    _DEFAULT_MEMORY_CPU_RATIO_WITH_GPU = 4  # Allocate more memory for GPU tasks
    _REPR = 'Kubernetes'
    _SINGLETON_REGION = 'kubernetes'
    _regions: List[clouds.Region] = [clouds.Region(_SINGLETON_REGION)]
    _CLOUD_UNSUPPORTED_FEATURES = {
        # TODO(romilb): Stopping might be possible to implement with
        #  container checkpointing introduced in Kubernetes v1.25. See:
        #  https://kubernetes.io/blog/2022/12/05/forensic-container-checkpointing-alpha/ # pylint: disable=line-too-long
        clouds.CloudImplementationFeatures.STOP: 'Kubernetes does not '
                                                 'support stopping VMs.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE: 'Spot instances are '
                                                          'not supported in '
                                                          'Kubernetes.',
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER: 'Custom disk '
                                                             'tiers are not '
                                                             'supported in '
                                                             'Kubernetes.',
    }

    IMAGE_CPU = 'skypilot:cpu-ubuntu-2004'
    IMAGE_GPU = 'skypilot:gpu-ubuntu-2004'

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        unsupported_features = cls._CLOUD_UNSUPPORTED_FEATURES
        curr_context = kubernetes_utils.get_current_kube_config_context_name()
        if curr_context == kubernetes_utils.KIND_CONTEXT_NAME:
            # If we are using KIND, the loadbalancer service will never be
            # assigned an external IP. Users may use ingress, but that requires
            # blocking HTTP port 80.
            # For now, we disable port opening feature on kind clusters.
            unsupported_features[
                clouds.CloudImplementationFeatures.OPEN_PORTS] = (
                    'Opening ports is not supported in Kubernetes when '
                    'using local kind cluster.')
        return unsupported_features

    @classmethod
    def regions(cls) -> List[clouds.Region]:
        return cls._regions

    @classmethod
    def regions_with_offering(cls, instance_type: Optional[str],
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        # No notion of regions in Kubernetes - return a single region.
        return cls.regions()

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        # TODO(romilb): Investigate how users can provide their own cost catalog
        #  for Kubernetes clusters.
        # For now, assume zero cost for Kubernetes clusters
        return 0.0

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        del accelerators, use_spot, region, zone  # unused
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    def __repr__(self):
        return self._REPR

    def is_same_cloud(self, other: clouds.Cloud) -> bool:
        return isinstance(other, Kubernetes)

    @classmethod
    def get_port(cls, svc_name) -> int:
        ns = kubernetes_utils.get_current_kube_config_context_namespace()
        return kubernetes_utils.get_port(svc_name, ns)

    @classmethod
    def get_default_instance_type(
            cls,
            cpus: Optional[str] = None,
            memory: Optional[str] = None,
            disk_tier: Optional[resources_utils.DiskTier] = None) -> str:
        # TODO(romilb): In the future, we may want to move the instance type
        #  selection + availability checking to a kubernetes_catalog module.
        del disk_tier  # Unused.
        # We strip '+' from resource requests since Kubernetes can provision
        # exactly the requested resources.
        instance_cpus = float(
            cpus.strip('+')) if cpus is not None else cls._DEFAULT_NUM_VCPUS
        instance_mem = float(memory.strip('+')) if memory is not None else \
            instance_cpus * cls._DEFAULT_MEMORY_CPU_RATIO
        virtual_instance_type = kubernetes_utils.KubernetesInstanceType(
            instance_cpus, instance_mem).name
        return virtual_instance_type

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, int]]:
        inst = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            instance_type)
        return {
            inst.accelerator_type: inst.accelerator_count
        } if (inst.accelerator_count is not None and
              inst.accelerator_type is not None) else None

    @classmethod
    def get_vcpus_mem_from_instance_type(
            cls, instance_type: str) -> Tuple[Optional[float], Optional[float]]:
        """Returns the #vCPUs and memory that the instance type offers."""
        k = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            instance_type)
        return k.cpus, k.memory

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[Optional[List[clouds.Zone]]]:
        del num_nodes, region, instance_type, accelerators, use_spot  # Unused.
        for r in cls.regions():
            yield r.zones

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> int:
        del image_id, region  # Unused.
        # We don't limit the image by its size compared to the disk size, as
        # we don't have a notion of disk size in Kubernetes.
        return 0

    def make_deploy_resources_variables(
            self,
            resources: 'resources_lib.Resources',
            cluster_name_on_cloud: str,
            region: Optional['clouds.Region'],
            zones: Optional[List['clouds.Zone']],
            dryrun: bool = False) -> Dict[str, Optional[str]]:
        del cluster_name_on_cloud, zones, dryrun  # Unused.
        if region is None:
            region = self._regions[0]

        r = resources
        acc_dict = self.get_accelerators_from_instance_type(r.instance_type)
        if acc_dict is not None:
            custom_resources = json.dumps(acc_dict, separators=(',', ':'))
        else:
            custom_resources = None

        # resources.memory and cpus are None if they are not explicitly set.
        # We fetch the default values for the instance type in that case.
        k = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            resources.instance_type)
        cpus = k.cpus
        mem = k.memory
        # Optionally populate accelerator information.
        acc_count = k.accelerator_count if k.accelerator_count else 0
        acc_type = k.accelerator_type if k.accelerator_type else None

        if resources.image_id is not None:
            # Use custom image specified in resources
            image_id_with_region = resources.image_id['kubernetes']
            image_id = image_id_with_region.lstrip('docker:')
        else:
            # Select image based on whether we are using GPUs or not.
            image_id = self.IMAGE_GPU if acc_count > 0 else self.IMAGE_CPU
            # Get the container image ID from the service catalog.
            image_id = service_catalog.get_image_id_from_tag(
                image_id, clouds='kubernetes')
        # TODO(romilb): Create a lightweight image for SSH jump host
        ssh_jump_image = service_catalog.get_image_id_from_tag(
            self.IMAGE_CPU, clouds='kubernetes')

        k8s_acc_label_key = None
        k8s_acc_label_value = None

        # If GPUs are requested, set node label to match the GPU type.
        if acc_count > 0 and acc_type is not None:
            k8s_acc_label_key, k8s_acc_label_value = \
                kubernetes_utils.get_gpu_label_key_value(acc_type)

        port_mode = network_utils.get_port_mode(None)

        deploy_vars = {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
            'cpus': str(cpus),
            'memory': str(mem),
            'accelerator_count': str(acc_count),
            'timeout': str(self.TIMEOUT),
            'k8s_namespace':
                kubernetes_utils.get_current_kube_config_context_namespace(),
            'k8s_port_mode': port_mode.value,
            'k8s_ssh_key_secret_name': self.SKY_SSH_KEY_SECRET_NAME,
            'k8s_acc_label_key': k8s_acc_label_key,
            'k8s_acc_label_value': k8s_acc_label_value,
            'k8s_ssh_jump_name': self.SKY_SSH_JUMP_NAME,
            'k8s_ssh_jump_image': ssh_jump_image,
            # TODO(romilb): Allow user to specify custom images
            'image_id': image_id,
        }

        return deploy_vars

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> Tuple[List['resources_lib.Resources'], List[str]]:
        fuzzy_candidate_list: List[str] = []
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return ([resources], fuzzy_candidate_list)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=Kubernetes(),
                    instance_type=instance_type,
                    accelerators=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators

        default_instance_type = Kubernetes.get_default_instance_type(
            cpus=resources.cpus,
            memory=resources.memory,
            disk_tier=resources.disk_tier)

        if accelerators is None:
            # For CPU only clusters, need no special handling
            chosen_instance_type = default_instance_type
        else:
            assert len(accelerators) == 1, resources
            # GPUs requested - build instance type.
            acc_type, acc_count = list(accelerators.items())[0]

            # Parse into KubernetesInstanceType
            k8s_instance_type = (kubernetes_utils.KubernetesInstanceType.
                                 from_instance_type(default_instance_type))

            gpu_task_cpus = k8s_instance_type.cpus
            # Special handling to bump up memory multiplier for GPU instances
            gpu_task_memory = (float(resources.memory.strip('+')) if
                               resources.memory is not None else gpu_task_cpus *
                               self._DEFAULT_MEMORY_CPU_RATIO_WITH_GPU)
            chosen_instance_type = (
                kubernetes_utils.KubernetesInstanceType.from_resources(
                    gpu_task_cpus, gpu_task_memory, acc_count, acc_type).name)

        # Check if requested instance type will fit in the cluster.
        # TODO(romilb): This will fail early for autoscaling clusters.
        fits, reason = kubernetes_utils.check_instance_fits(
            chosen_instance_type)
        if not fits:
            logger.debug(f'Instance type {chosen_instance_type} does '
                         'not fit in the Kubernetes cluster. '
                         f'Reason: {reason}')
            return [], []

        # No fuzzy lists for Kubernetes
        return _make([chosen_instance_type]), []

    @classmethod
    def check_credentials(cls) -> Tuple[bool, Optional[str]]:
        if os.path.exists(os.path.expanduser(CREDENTIAL_PATH)):
            # Test using python API
            try:
                return kubernetes_utils.check_credentials()
            except Exception as e:  # pylint: disable=broad-except
                return (False, 'Credential check failed: '
                        f'{common_utils.format_exception(e)}')
        else:
            return (False, 'Credentials not found - '
                    f'check if {CREDENTIAL_PATH} exists.')

    def get_credential_file_mounts(self) -> Dict[str, str]:
        # Upload kubeconfig to the default path to avoid having to set
        # KUBECONFIG in the environment.
        return {DEFAULT_KUBECONFIG_PATH: CREDENTIAL_PATH}

    def instance_type_exists(self, instance_type: str) -> bool:
        return kubernetes_utils.KubernetesInstanceType.is_valid_instance_type(
            instance_type)

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        if region != self._SINGLETON_REGION:
            raise ValueError(
                'Kubernetes support does not support setting region.'
                ' Cluster used is determined by the kubeconfig.')
        if zone is not None:
            raise ValueError('Kubernetes support does not support setting zone.'
                             ' Cluster used is determined by the kubeconfig.')
        return region, zone

    @classmethod
    def get_current_user_identity(cls) -> Optional[List[str]]:
        k8s = kubernetes.get_kubernetes()
        try:
            _, current_context = k8s.config.list_kube_config_contexts()
            if 'namespace' in current_context['context']:
                namespace = current_context['context']['namespace']
            else:
                namespace = kubernetes_utils.DEFAULT_NAMESPACE

            user = current_context['context']['user']
            cluster = current_context['context']['cluster']
            return [f'{cluster}_{user}_{namespace}']
        except k8s.config.config_exception.ConfigException:
            return None



"""Script to label GPU nodes in a Kubernetes cluster for use with SkyPilot"""
import argparse
import hashlib
import os
import subprocess
from typing import Tuple

from kubernetes import client
from kubernetes import config
import yaml

import sky
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import rich_utils


def prerequisite_check() -> Tuple[bool, str]:
    """Checks if kubectl is installed and kubeconfig is set up"""
    reason = ''
    prereq_ok = False
    try:
        subprocess.run(['kubectl', 'get', 'pods'],
                       check=True,
                       capture_output=True)
        prereq_ok = True
    except FileNotFoundError:
        reason = 'kubectl not found. Please install kubectl and try again.'
    except subprocess.CalledProcessError as e:
        output = e.output.decode('utf-8')
        reason = 'Error running kubectl: ' + output
    return prereq_ok, reason


def cleanup() -> Tuple[bool, str]:
    """Deletes all Kubernetes resources created by this script

    Used to provide idempotency when the script is run multiple times. Also
    invoked if --cleanup is passed to the script.
    """
    # Delete any existing GPU labeler Kubernetes resources:
    del_command = ('kubectl delete pods,services,deployments,jobs,daemonsets,'
                   'replicasets,configmaps,secrets,pv,pvc,clusterrole,'
                   'serviceaccount,clusterrolebinding -n kube-system '
                   '-l job=sky-gpu-labeler')

    success = False
    reason = ''
    with rich_utils.safe_status('Cleaning up existing GPU labeling '
                                'resources'):
        try:
            subprocess.run(del_command.split(), check=True, capture_output=True)
            success = True
        except subprocess.CalledProcessError as e:
            output = e.output.decode('utf-8')
            reason = 'Error deleting existing GPU labeler resources: ' + output
        return success, reason


def get_node_hash(node_name: str):
    # Generates a 32 character md5 hash from a string
    md5_hash = hashlib.md5(node_name.encode()).hexdigest()
    return md5_hash[:32]


def label():
    deletion_success, reason = cleanup()
    if not deletion_success:
        print(reason)
        return

    sky_dir = os.path.dirname(sky.__file__)
    manifest_dir = os.path.join(sky_dir, 'utils/kubernetes')

    # Apply the RBAC manifest using kubectl since it contains multiple resources
    with rich_utils.safe_status('Setting up GPU labeling'):
        rbac_manifest_path = os.path.join(manifest_dir,
                                          'k8s_gpu_labeler_setup.yaml')
        try:
            subprocess.check_output(
                ['kubectl', 'apply', '-f', rbac_manifest_path])
        except subprocess.CalledProcessError as e:
            output = e.output.decode('utf-8')
            print('Error setting up GPU labeling: ' + output)
            return

    with rich_utils.safe_status('Creating GPU labeler jobs'):
        config.load_kube_config()

        v1 = client.CoreV1Api()
        batch_v1 = client.BatchV1Api()
        # Load the job manifest
        job_manifest_path = os.path.join(manifest_dir,
                                         'k8s_gpu_labeler_job.yaml')

        with open(job_manifest_path, 'r', encoding='utf-8') as file:
            job_manifest = yaml.safe_load(file)

        # Iterate over nodes
        nodes = v1.list_node().items

        # Get the list of nodes with GPUs
        gpu_nodes = []
        for node in nodes:
            if 'nvidia.com/gpu' in node.status.capacity:
                gpu_nodes.append(node)

        print(f'Found {len(gpu_nodes)} GPU nodes in the cluster')

        # Check if the 'nvidia' RuntimeClass exists
        try:
            nvidia_exists = kubernetes_utils.check_nvidia_runtime_class()
        except Exception as e:  # pylint: disable=broad-except
            print('Error occurred while checking for nvidia RuntimeClass: '
                  f'{str(e)}')
            print('Continuing without using nvidia RuntimeClass. '
                  'This may fail on K3s clusters. '
                  'For more details, refer to K3s deployment notes at: '
                  'https://skypilot.readthedocs.io/en/latest/reference/kubernetes/kubernetes-setup.html')  # pylint: disable=line-too-long
            nvidia_exists = False

        if nvidia_exists:
            print('Using nvidia RuntimeClass for GPU labeling.')
            job_manifest['spec']['template']['spec'][
                'runtimeClassName'] = 'nvidia'
        else:
            print('Using default RuntimeClass for GPU labeling.')

        for node in gpu_nodes:
            node_name = node.metadata.name

            # Modify the job manifest for the current node
            job_manifest['metadata']['name'] = ('sky-gpu-labeler-'
                                                f'{get_node_hash(node_name)}')
            job_manifest['spec']['template']['spec']['nodeSelector'] = {
                'kubernetes.io/hostname': node_name
            }
            namespace = job_manifest['metadata']['namespace']

            # Create the job for this node`
            batch_v1.create_namespaced_job(namespace, job_manifest)
            print(f'Created GPU labeler job for node {node_name}')
    if len(gpu_nodes) == 0:
        print('No GPU nodes found in the cluster. If you have GPU nodes, '
              'please ensure that they have the label '
              '`nvidia.com/gpu: <number of GPUs>`')
    else:
        print('GPU labeling started - this may take a few minutes to complete.'
              '\nTo check the status of GPU labeling jobs, run '
              '`kubectl get jobs -n kube-system -l job=sky-gpu-labeler`'
              '\nYou can check if nodes have been labeled by running '
              '`kubectl describe nodes` and looking for labels of the format '
              '`skypilot.co/accelerators: <gpu_name>`. ')


def main():
    parser = argparse.ArgumentParser(
        description='Labels GPU nodes in a Kubernetes cluster for use with '
        'SkyPilot. Operates by running a job on each node that '
        'parses nvidia-smi and patches the node with new labels. '
        'Labels created are of the format '
        'skypilot.co/accelerators: <gpu_name>. Automatically '
        'creates a service account and cluster role binding with '
        'permissions to list nodes and create labels.')
    parser.add_argument('--cleanup',
                        action='store_true',
                        help='delete all GPU labeler resources in the '
                        'Kubernetes cluster.')
    args = parser.parse_args()

    # Check if kubectl is installed and kubeconfig is set up
    prereq_ok, reason = prerequisite_check()
    if not prereq_ok:
        print(reason)
        return

    if args.cleanup:
        cleanup()
    else:
        label()


if __name__ == '__main__':
    main()



"""User interfaces with managed spot jobs."""
import collections
import enum
import json
import os
import pathlib
import shlex
import time
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

import colorama
import filelock
from typing_extensions import Literal

from sky import backends
from sky import exceptions
from sky import global_user_state
from sky import sky_logging
from sky.backends import backend_utils
from sky.skylet import constants
from sky.skylet import job_lib
from sky.skylet.log_lib import run_bash_command_with_log
from sky.spot import constants as spot_constants
from sky.spot import spot_state
from sky.utils import common_utils
from sky.utils import log_utils
from sky.utils import rich_utils
from sky.utils import subprocess_utils

if typing.TYPE_CHECKING:
    import sky
    from sky import dag as dag_lib

logger = sky_logging.init_logger(__name__)

# Add user hash so that two users don't have the same controller VM on
# shared-account clouds such as GCP.
SPOT_CONTROLLER_NAME: str = (
    f'sky-spot-controller-{common_utils.get_user_hash()}')
SIGNAL_FILE_PREFIX = '/tmp/sky_spot_controller_signal_{}'
# Controller checks its job's status every this many seconds.
JOB_STATUS_CHECK_GAP_SECONDS = 20

# Controller checks if its job has started every this many seconds.
JOB_STARTED_STATUS_CHECK_GAP_SECONDS = 5

_SPOT_STATUS_CACHE = '~/.sky/spot_status_cache.txt'

_LOG_STREAM_CHECK_CONTROLLER_GAP_SECONDS = 5

_JOB_WAITING_STATUS_MESSAGE = ('[bold cyan]Waiting for the task to start'
                               '{status_str}.[/] It may take a few minutes.')
_JOB_CANCELLED_MESSAGE = (
    '[bold cyan]Waiting for the task status to be updated.'
    '[/] It may take a minute.')

# The maximum time to wait for the spot job status to transition to terminal
# state, after the job finished. This is a safeguard to avoid the case where
# the spot job status fails to be updated and keep the `sky spot logs` blocking
# for a long time.
_FINAL_SPOT_STATUS_WAIT_TIMEOUT_SECONDS = 20


class UserSignal(enum.Enum):
    """The signal to be sent to the user."""
    CANCEL = 'CANCEL'
    # NOTE: We can have more communication signals here if needed
    # in the future.


# ====== internal functions ======
def get_job_status(backend: 'backends.CloudVmRayBackend',
                   cluster_name: str) -> Optional['job_lib.JobStatus']:
    """Check the status of the job running on the spot cluster.

    It can be None, INIT, RUNNING, SUCCEEDED, FAILED, FAILED_SETUP or CANCELLED.
    """
    handle = global_user_state.get_handle_from_cluster_name(cluster_name)
    assert isinstance(handle, backends.CloudVmRayResourceHandle), handle
    status = None
    try:
        logger.info('=== Checking the job status... ===')
        statuses = backend.get_job_status(handle, stream_logs=False)
        status = list(statuses.values())[0]
        if status is None:
            logger.info('No job found.')
        else:
            logger.info(f'Job status: {status}')
    except exceptions.CommandError:
        logger.info('Failed to connect to the cluster.')
    logger.info('=' * 34)
    return status


def update_spot_job_status(job_id: Optional[int] = None):
    """Update spot job status if the controller job failed abnormally.

    Check the status of the controller job. If it is not running, it must have
    exited abnormally, and we should set the job status to FAILED_CONTROLLER.
    `end_at` will be set to the current timestamp for the job when above
    happens, which could be not accurate based on the frequency this function
    is called.
    """
    if job_id is None:
        job_ids = spot_state.get_nonterminal_job_ids_by_name(None)
    else:
        job_ids = [job_id]
    for job_id_ in job_ids:
        controller_status = job_lib.get_status(job_id_)
        if controller_status is None or controller_status.is_terminal():
            logger.error(f'Controller for job {job_id_} has exited abnormally. '
                         'Setting the job status to FAILED_CONTROLLER.')
            tasks = spot_state.get_spot_jobs(job_id_)
            for task in tasks:
                task_name = task['job_name']
                # Tear down the abnormal spot cluster to avoid resource leakage.
                cluster_name = generate_spot_cluster_name(task_name, job_id_)
                handle = global_user_state.get_handle_from_cluster_name(
                    cluster_name)
                if handle is not None:
                    backend = backend_utils.get_backend_from_handle(handle)
                    max_retry = 3
                    for retry_cnt in range(max_retry):
                        try:
                            backend.teardown(handle, terminate=True)
                            break
                        except RuntimeError:
                            logger.error('Failed to tear down the spot cluster '
                                         f'{cluster_name!r}. Retrying '
                                         f'[{retry_cnt}/{max_retry}].')

            # The controller job for this spot job is not running: it must
            # have exited abnormally, and we should set the job status to
            # FAILED_CONTROLLER.
            # The `set_failed` will only update the task's status if the
            # status is non-terminal.
            spot_state.set_failed(
                job_id_,
                task_id=None,
                failure_type=spot_state.SpotStatus.FAILED_CONTROLLER,
                failure_reason=
                'Controller process has exited abnormally. For more details,'
                f' run: sky spot logs --controller {job_id_}')


def get_job_timestamp(backend: 'backends.CloudVmRayBackend', cluster_name: str,
                      get_end_time: bool) -> float:
    """Get the submitted/ended time of the job."""
    code = job_lib.JobLibCodeGen.get_job_submitted_or_ended_timestamp_payload(
        job_id=None, get_ended_time=get_end_time)
    handle = global_user_state.get_handle_from_cluster_name(cluster_name)
    returncode, stdout, stderr = backend.run_on_head(handle,
                                                     code,
                                                     stream_logs=False,
                                                     require_outputs=True)
    subprocess_utils.handle_returncode(returncode, code,
                                       'Failed to get job time.',
                                       stdout + stderr)
    stdout = common_utils.decode_payload(stdout)
    return float(stdout)


def event_callback_func(job_id: int, task_id: int, task: 'sky.Task'):
    """Run event callback for the task."""

    def callback_func(state: str):
        event_callback = task.event_callback if task else None
        if event_callback is None or task is None:
            return
        event_callback = event_callback.strip()
        cluster_name = generate_spot_cluster_name(task.name,
                                                  job_id) if task.name else None
        logger.info(f'=== START: event callback for {state!r} ===')
        log_path = os.path.join(constants.SKY_LOGS_DIRECTORY, 'spot_event',
                                f'spot-callback-{job_id}-{task_id}.log')
        result = run_bash_command_with_log(
            bash_command=event_callback,
            log_path=log_path,
            env_vars=dict(
                SKYPILOT_JOB_ID=str(
                    task.envs.get(constants.TASK_ID_ENV_VAR_DEPRECATED,
                                  'N.A.')),
                SKYPILOT_TASK_ID=str(
                    task.envs.get(constants.TASK_ID_ENV_VAR, 'N.A.')),
                SKYPILOT_TASK_IDS=str(
                    task.envs.get(constants.TASK_ID_LIST_ENV_VAR, 'N.A.')),
                TASK_ID=str(task_id),
                JOB_ID=str(job_id),
                JOB_STATUS=state,
                CLUSTER_NAME=cluster_name or '',
                TASK_NAME=task.name or '',
                # TODO(MaoZiming): Future event type Job or Spot.
                EVENT_TYPE='Spot'))
        logger.info(
            f'Bash:{event_callback},log_path:{log_path},result:{result}')
        logger.info(f'=== END: event callback for {state!r} ===')

    return callback_func


# ======== user functions ========


def generate_spot_cluster_name(task_name: str, job_id: int) -> str:
    """Generate spot cluster name."""
    # Truncate the task name to 30 chars to avoid the cluster name being too
    # long after appending the job id, which will cause another truncation in
    # the underlying sky.launch, hiding the `job_id` in the cluster name.
    cluster_name = common_utils.make_cluster_name_on_cloud(
        task_name,
        spot_constants.SPOT_CLUSTER_NAME_PREFIX_LENGTH,
        add_user_hash=False)
    return f'{cluster_name}-{job_id}'


def cancel_jobs_by_id(job_ids: Optional[List[int]]) -> str:
    """Cancel jobs by id.

    If job_ids is None, cancel all jobs.
    """
    if job_ids is None:
        job_ids = spot_state.get_nonterminal_job_ids_by_name(None)
    job_ids = list(set(job_ids))
    if len(job_ids) == 0:
        return 'No job to cancel.'
    job_id_str = ', '.join(map(str, job_ids))
    logger.info(f'Cancelling jobs {job_id_str}.')
    cancelled_job_ids = []
    for job_id in job_ids:
        # Check the status of the managed spot job status. If it is in
        # terminal state, we can safely skip it.
        job_status = spot_state.get_status(job_id)
        if job_status is None:
            logger.info(f'Job {job_id} not found. Skipped.')
            continue
        elif job_status.is_terminal():
            logger.info(f'Job {job_id} is already in terminal state '
                        f'{job_status.value}. Skipped.')
            continue

        update_spot_job_status(job_id)

        # Send the signal to the spot job controller.
        signal_file = pathlib.Path(SIGNAL_FILE_PREFIX.format(job_id))
        # Filelock is needed to prevent race condition between signal
        # check/removal and signal writing.
        # TODO(mraheja): remove pylint disabling when filelock version updated
        # pylint: disable=abstract-class-instantiated
        with filelock.FileLock(str(signal_file) + '.lock'):
            with signal_file.open('w', encoding='utf-8') as f:
                f.write(UserSignal.CANCEL.value)
                f.flush()
        cancelled_job_ids.append(job_id)

    if len(cancelled_job_ids) == 0:
        return 'No job to cancel.'
    identity_str = f'Job with ID {cancelled_job_ids[0]} is'
    if len(cancelled_job_ids) > 1:
        cancelled_job_ids_str = ', '.join(map(str, cancelled_job_ids))
        identity_str = f'Jobs with IDs {cancelled_job_ids_str} are'

    return f'{identity_str} scheduled to be cancelled.'


def cancel_job_by_name(job_name: str) -> str:
    """Cancel a job by name."""
    job_ids = spot_state.get_nonterminal_job_ids_by_name(job_name)
    if len(job_ids) == 0:
        return f'No running job found with name {job_name!r}.'
    if len(job_ids) > 1:
        return (f'{colorama.Fore.RED}Multiple running jobs found '
                f'with name {job_name!r}.\n'
                f'Job IDs: {job_ids}{colorama.Style.RESET_ALL}')
    cancel_jobs_by_id(job_ids)
    return f'Job {job_name!r} is scheduled to be cancelled.'


def stream_logs_by_id(job_id: int, follow: bool = True) -> str:
    """Stream logs by job id."""
    controller_status = job_lib.get_status(job_id)
    status_msg = ('[bold cyan]Waiting for controller process to be RUNNING'
                  '{status_str}[/].')
    status_display = rich_utils.safe_status(status_msg.format(status_str=''))
    num_tasks = spot_state.get_num_tasks(job_id)

    with status_display:
        prev_msg = None
        while (controller_status != job_lib.JobStatus.RUNNING and
               (controller_status is None or
                not controller_status.is_terminal())):
            status_str = 'None'
            if controller_status is not None:
                status_str = controller_status.value
            msg = status_msg.format(status_str=f' (status: {status_str})')
            if msg != prev_msg:
                status_display.update(msg)
                prev_msg = msg
            time.sleep(_LOG_STREAM_CHECK_CONTROLLER_GAP_SECONDS)
            controller_status = job_lib.get_status(job_id)

        msg = _JOB_WAITING_STATUS_MESSAGE.format(status_str='')
        status_display.update(msg)
        prev_msg = msg
        spot_job_status = spot_state.get_status(job_id)
        while spot_job_status is None:
            time.sleep(1)
            spot_job_status = spot_state.get_status(job_id)

        if spot_job_status.is_terminal():
            job_msg = ''
            if spot_job_status.is_failed():
                job_msg = (
                    f'\nFailure reason: {spot_state.get_failure_reason(job_id)}'
                )
            return (f'{colorama.Fore.YELLOW}'
                    f'Job {job_id} is already in terminal state '
                    f'{spot_job_status.value}. Logs will not be shown.'
                    f'{colorama.Style.RESET_ALL}{job_msg}')
        backend = backends.CloudVmRayBackend()
        task_id, spot_status = spot_state.get_latest_task_id_status(job_id)

        # task_id and spot_status can be None if the controller process just
        # started and the spot status has not set to PENDING yet.
        while spot_status is None or not spot_status.is_terminal():
            handle = None
            if task_id is not None:
                task_name = spot_state.get_task_name(job_id, task_id)
                cluster_name = generate_spot_cluster_name(task_name, job_id)
                handle = global_user_state.get_handle_from_cluster_name(
                    cluster_name)

            # Check the handle: The cluster can be preempted and removed from
            # the table before the spot state is updated by the controller. In
            # this case, we should skip the logging, and wait for the next
            # round of status check.
            if handle is None or spot_status != spot_state.SpotStatus.RUNNING:
                status_str = ''
                if (spot_status is not None and
                        spot_status != spot_state.SpotStatus.RUNNING):
                    status_str = f' (status: {spot_status.value})'
                logger.debug(
                    f'INFO: The log is not ready yet{status_str}. '
                    f'Waiting for {JOB_STATUS_CHECK_GAP_SECONDS} seconds.')
                msg = _JOB_WAITING_STATUS_MESSAGE.format(status_str=status_str)
                if msg != prev_msg:
                    status_display.update(msg)
                    prev_msg = msg
                time.sleep(JOB_STATUS_CHECK_GAP_SECONDS)
                task_id, spot_status = (
                    spot_state.get_latest_task_id_status(job_id))
                continue
            assert spot_status is not None
            assert isinstance(handle, backends.CloudVmRayResourceHandle), handle
            status_display.stop()
            returncode = backend.tail_logs(handle,
                                           job_id=None,
                                           spot_job_id=job_id,
                                           follow=follow)
            if returncode == 0:
                # If the log tailing exit successfully (the real job can be
                # SUCCEEDED or FAILED), we can safely break the loop. We use the
                # status in job queue to show the information, as the spot_state
                # is not updated yet.
                job_statuses = backend.get_job_status(handle, stream_logs=False)
                job_status = list(job_statuses.values())[0]
                assert job_status is not None, 'No job found.'
                if job_status != job_lib.JobStatus.CANCELLED:
                    assert task_id is not None, job_id
                    if task_id < num_tasks - 1 and follow:
                        # The log for the current job is finished. We need to
                        # wait until next job to be started.
                        logger.debug(
                            f'INFO: Log for the current task ({task_id}) '
                            'is finished. Waiting for the next task\'s log '
                            'to be started.')
                        status_display.update('Waiting for the next task: '
                                              f'{task_id + 1}.')
                        status_display.start()
                        original_task_id = task_id
                        while True:
                            task_id, spot_status = (
                                spot_state.get_latest_task_id_status(job_id))
                            if original_task_id != task_id:
                                break
                            time.sleep(JOB_STATUS_CHECK_GAP_SECONDS)
                        continue
                    else:
                        break
                # The job can be cancelled by the user or the controller (when
                # the cluster is partially preempted).
                logger.debug(
                    'INFO: Job is cancelled. Waiting for the status update in '
                    f'{JOB_STATUS_CHECK_GAP_SECONDS} seconds.')
            else:
                logger.debug(
                    f'INFO: (Log streaming) Got return code {returncode}. '
                    f'Retrying in {JOB_STATUS_CHECK_GAP_SECONDS} seconds.')
            # Finish early if the spot status is already in terminal state.
            spot_status = spot_state.get_status(job_id)
            assert spot_status is not None, job_id
            if spot_status.is_terminal():
                break
            logger.info(f'{colorama.Fore.YELLOW}The job is preempted.'
                        f'{colorama.Style.RESET_ALL}')
            msg = _JOB_CANCELLED_MESSAGE
            status_display.update(msg)
            prev_msg = msg
            status_display.start()
            # If the tailing fails, it is likely that the cluster fails, so we
            # wait a while to make sure the spot state is updated by the
            # controller, and check the spot queue again.
            # Wait a bit longer than the controller, so as to make sure the
            # spot state is updated.
            time.sleep(3 * JOB_STATUS_CHECK_GAP_SECONDS)
            spot_status = spot_state.get_status(job_id)

    # The spot_status may not be in terminal status yet, since the controllerhas
    # not updated the spot state yet. We wait for a while, until the spot state
    # is updated.
    wait_seconds = 0
    spot_status = spot_state.get_status(job_id)
    assert spot_status is not None, job_id
    while (not spot_status.is_terminal() and follow and
           wait_seconds < _FINAL_SPOT_STATUS_WAIT_TIMEOUT_SECONDS):
        time.sleep(1)
        wait_seconds += 1
        spot_status = spot_state.get_status(job_id)
        assert spot_status is not None, job_id

    logger.info(f'Logs finished for job {job_id} '
                f'(status: {spot_status.value}).')
    return ''


def stream_logs_by_name(job_name: str, follow: bool = True) -> str:
    """Stream logs by name."""
    job_ids = spot_state.get_nonterminal_job_ids_by_name(job_name)
    if len(job_ids) == 0:
        return (f'{colorama.Fore.RED}No job found with name {job_name!r}.'
                f'{colorama.Style.RESET_ALL}')
    if len(job_ids) > 1:
        return (f'{colorama.Fore.RED}Multiple running jobs found '
                f'with name {job_name!r}.\n'
                f'Job IDs: {job_ids}{colorama.Style.RESET_ALL}')
    stream_logs_by_id(job_ids[0], follow)
    return ''


def dump_spot_job_queue() -> str:
    jobs = spot_state.get_spot_jobs()

    for job in jobs:
        end_at = job['end_at']
        if end_at is None:
            end_at = time.time()

        job_submitted_at = job['last_recovered_at'] - job['job_duration']
        if job['status'] == spot_state.SpotStatus.RECOVERING:
            # When job is recovering, the duration is exact job['job_duration']
            job_duration = job['job_duration']
        elif job_submitted_at > 0:
            job_duration = end_at - job_submitted_at
        else:
            # When job_start_at <= 0, that means the last_recovered_at is not
            # set yet, i.e. the job is not started.
            job_duration = 0
        job['job_duration'] = job_duration
        job['status'] = job['status'].value

        cluster_name = generate_spot_cluster_name(job['task_name'],
                                                  job['job_id'])
        handle = global_user_state.get_handle_from_cluster_name(cluster_name)
        if handle is not None:
            assert isinstance(handle, backends.CloudVmRayResourceHandle)
            job['cluster_resources'] = (
                f'{handle.launched_nodes}x {handle.launched_resources}')
            job['region'] = handle.launched_resources.region
        else:
            # FIXME(zongheng): display the last cached values for these.
            job['cluster_resources'] = '-'
            job['region'] = '-'

    return common_utils.encode_payload(jobs)


def load_spot_job_queue(payload: str) -> List[Dict[str, Any]]:
    """Load job queue from json string."""
    jobs = common_utils.decode_payload(payload)
    for job in jobs:
        job['status'] = spot_state.SpotStatus(job['status'])
    return jobs


@typing.overload
def format_job_table(tasks: List[Dict[str, Any]],
                     show_all: bool,
                     return_rows: Literal[False] = False,
                     max_jobs: Optional[int] = None) -> str:
    ...


@typing.overload
def format_job_table(tasks: List[Dict[str, Any]],
                     show_all: bool,
                     return_rows: Literal[True],
                     max_jobs: Optional[int] = None) -> List[List[str]]:
    ...


def format_job_table(
        tasks: List[Dict[str, Any]],
        show_all: bool,
        return_rows: bool = False,
        max_jobs: Optional[int] = None) -> Union[str, List[List[str]]]:
    """Returns spot jobs as a formatted string.

    Args:
        jobs: A list of spot jobs.
        show_all: Whether to show all columns.
        max_jobs: The maximum number of jobs to show in the table.
        return_rows: If True, return the rows as a list of strings instead of
          all rows concatenated into a single string.

    Returns: A formatted string of spot jobs, if not `return_rows`; otherwise a
      list of "rows" (each of which is a list of str).
    """
    columns = [
        'ID', 'TASK', 'NAME', 'RESOURCES', 'SUBMITTED', 'TOT. DURATION',
        'JOB DURATION', '#RECOVERIES', 'STATUS'
    ]
    if show_all:
        columns += ['STARTED', 'CLUSTER', 'REGION', 'FAILURE']
    job_table = log_utils.create_table(columns)

    status_counts: Dict[str, int] = collections.defaultdict(int)
    for task in tasks:
        if not task['status'].is_terminal():
            status_counts[task['status'].value] += 1

    all_tasks = tasks
    if max_jobs is not None:
        all_tasks = tasks[:max_jobs]
    jobs = collections.defaultdict(list)
    for task in all_tasks:
        # The tasks within the same job_id are already sorted
        # by the task_id.
        jobs[task['job_id']].append(task)

    for job_id, job_tasks in jobs.items():
        if len(job_tasks) > 1:
            # Aggregate the tasks into a new row in the table.
            job_name = job_tasks[0]['job_name']
            job_duration = 0
            submitted_at = None
            end_at: Optional[int] = 0
            recovery_cnt = 0
            spot_status = spot_state.SpotStatus.SUCCEEDED
            failure_reason = None
            current_task_id = len(job_tasks) - 1
            for task in job_tasks:
                job_duration += task['job_duration']
                if task['submitted_at'] is not None:
                    if (submitted_at is None or
                            submitted_at > task['submitted_at']):
                        submitted_at = task['submitted_at']
                if task['end_at'] is not None:
                    if end_at is not None and end_at < task['end_at']:
                        end_at = task['end_at']
                else:
                    end_at = None
                recovery_cnt += task['recovery_count']
                if spot_status == spot_state.SpotStatus.SUCCEEDED:
                    # Use the first non-succeeded status.
                    # TODO(zhwu): we should not blindly use the first non-
                    # succeeded as the status could be changed to SUBMITTED
                    # when going from one task to the next one, which can be
                    # confusing.
                    spot_status = task['status']
                    current_task_id = task['task_id']

                if (failure_reason is None and
                        task['status'] > spot_state.SpotStatus.SUCCEEDED):
                    failure_reason = task['failure_reason']

            job_duration = log_utils.readable_time_duration(0,
                                                            job_duration,
                                                            absolute=True)
            submitted = log_utils.readable_time_duration(submitted_at)
            total_duration = log_utils.readable_time_duration(submitted_at,
                                                              end_at,
                                                              absolute=True)

            status_str = spot_status.colored_str()
            if (spot_status < spot_state.SpotStatus.RUNNING and
                    current_task_id > 0):
                status_str += f' (task: {current_task_id})'

            job_values = [
                job_id,
                '',
                job_name,
                '-',
                submitted,
                total_duration,
                job_duration,
                recovery_cnt,
                status_str,
            ]
            if show_all:
                job_values.extend([
                    '-',
                    '-',
                    '-',
                    failure_reason if failure_reason is not None else '-',
                ])
            job_table.add_row(job_values)

        for task in job_tasks:
            # The job['job_duration'] is already calculated in
            # dump_spot_job_queue().
            job_duration = log_utils.readable_time_duration(
                0, task['job_duration'], absolute=True)
            submitted = log_utils.readable_time_duration(task['submitted_at'])
            values = [
                task['job_id'] if len(job_tasks) == 1 else ' \u21B3',
                task['task_id'] if len(job_tasks) > 1 else '-',
                task['task_name'],
                task['resources'],
                # SUBMITTED
                submitted if submitted != '-' else submitted,
                # TOT. DURATION
                log_utils.readable_time_duration(task['submitted_at'],
                                                 task['end_at'],
                                                 absolute=True),
                job_duration,
                task['recovery_count'],
                task['status'].colored_str(),
            ]
            if show_all:
                values.extend([
                    # STARTED
                    log_utils.readable_time_duration(task['start_at']),
                    task['cluster_resources'],
                    task['region'],
                    task['failure_reason']
                    if task['failure_reason'] is not None else '-',
                ])
            job_table.add_row(values)

        if len(job_tasks) > 1:
            # Add a row to separate the aggregated job from the next job.
            job_table.add_row([''] * len(columns))
    status_str = ', '.join([
        f'{count} {status}' for status, count in sorted(status_counts.items())
    ])
    if status_str:
        status_str = f'In progress tasks: {status_str}'
    else:
        status_str = 'No in progress tasks.'
    output = status_str
    if str(job_table):
        output += f'\n{job_table}'
    if return_rows:
        return job_table.rows
    return output


class SpotCodeGen:
    """Code generator for managed spot job utility functions.

    Usage:

      >> codegen = SpotCodegen.show_jobs(...)
    """
    _PREFIX = [
        'from sky.spot import spot_state',
        'from sky.spot import spot_utils',
    ]

    @classmethod
    def get_job_table(cls) -> str:
        code = [
            'job_table = spot_utils.dump_spot_job_queue()',
            'print(job_table, flush=True)',
        ]
        return cls._build(code)

    @classmethod
    def cancel_jobs_by_id(cls, job_ids: Optional[List[int]]) -> str:
        code = [
            f'msg = spot_utils.cancel_jobs_by_id({job_ids})',
            'print(msg, end="", flush=True)',
        ]
        return cls._build(code)

    @classmethod
    def cancel_job_by_name(cls, job_name: str) -> str:
        code = [
            f'msg = spot_utils.cancel_job_by_name({job_name!r})',
            'print(msg, end="", flush=True)',
        ]
        return cls._build(code)

    @classmethod
    def stream_logs_by_name(cls, job_name: str, follow: bool = True) -> str:
        code = [
            f'msg = spot_utils.stream_logs_by_name({job_name!r}, '
            f'follow={follow})',
            'print(msg, flush=True)',
        ]
        return cls._build(code)

    @classmethod
    def stream_logs_by_id(cls,
                          job_id: Optional[int],
                          follow: bool = True) -> str:
        code = [
            f'job_id = {job_id} if {job_id} is not None '
            'else spot_state.get_latest_job_id()',
            f'msg = spot_utils.stream_logs_by_id(job_id, follow={follow})',
            'print(msg, flush=True)',
        ]
        return cls._build(code)

    @classmethod
    def set_pending(cls, job_id: int, spot_dag: 'dag_lib.Dag') -> str:
        dag_name = spot_dag.name
        # Add the spot job to spot queue table.
        code = [
            f'spot_state.set_job_name('
            f'{job_id}, {dag_name!r})',
        ]
        for task_id, task in enumerate(spot_dag.tasks):
            resources_str = backend_utils.get_task_resources_str(task)
            code += [
                f'spot_state.set_pending('
                f'{job_id}, {task_id}, {task.name!r}, '
                f'{resources_str!r})',
            ]
        return cls._build(code)

    @classmethod
    def _build(cls, code: List[str]) -> str:
        code = cls._PREFIX + code
        generated_code = '; '.join(code)
        return f'python3 -u -c {shlex.quote(generated_code)}'


def dump_job_table_cache(job_table: str):
    """Dump job table cache to file."""
    cache_file = pathlib.Path(_SPOT_STATUS_CACHE).expanduser()
    with cache_file.open('w', encoding='utf-8') as f:
        json.dump((time.time(), job_table), f)


def load_job_table_cache() -> Optional[Tuple[float, str]]:
    """Load job table cache from file.

    Returns:
        A tuple of (timestamp, job_table), where the timestamp is
        the time when the job table is dumped and the job_table is
        the dumped job table in string.
        None if the cache file does not exist.
    """
    cache_file = pathlib.Path(_SPOT_STATUS_CACHE).expanduser()
    if not cache_file.exists():
        return None
    with cache_file.open('r', encoding='utf-8') as f:
        return json.load(f)



import argparse
import time

import requests


def send_get_request(ip, port):
    url = f"http://{ip}:{port}"
    try:
        response = requests.get(url)
        print(
            f"GET request sent to {url} | Status code: {response.status_code}")
    except requests.RequestException as e:
        print(f"Error sending GET request to {url}: {e}")


def main():
    parser = argparse.ArgumentParser(
        description="Send GET requests to a specified IP address and port.")
    parser.add_argument("--ip-address",
                        help="The IP address to send requests to.")
    parser.add_argument("--port", type=int, help="The port number to use.")
    parser.add_argument(
        "--frequency",
        type=float,
        default=5,
        help="Sending frequency in seconds (default is 5 seconds).")

    args = parser.parse_args()

    ip_address = args.ip_address
    port = args.port
    sending_frequency_seconds = args.frequency

    while True:
        send_get_request(ip_address, port)
        time.sleep(sending_frequency_seconds)


if __name__ == "__main__":
    main()



"""The database for services information."""
import collections
import enum
import pathlib
import pickle
import sqlite3
import typing
from typing import Any, Dict, List, Optional

import colorama

from sky.serve import constants
from sky.utils import db_utils

if typing.TYPE_CHECKING:
    import sky
    from sky.serve import replica_managers
    from sky.serve import service_spec

_DB_PATH = pathlib.Path(constants.SKYSERVE_METADATA_DIR) / 'services.db'
_DB_PATH = _DB_PATH.expanduser().absolute()
_DB_PATH.parents[0].mkdir(parents=True, exist_ok=True)
_DB_PATH = str(_DB_PATH)


def create_table(cursor: 'sqlite3.Cursor', conn: 'sqlite3.Connection') -> None:
    """Creates the service and replica tables if they do not exist."""

    # auto_restart column is deprecated.
    cursor.execute("""\
        CREATE TABLE IF NOT EXISTS services (
        name TEXT PRIMARY KEY,
        controller_job_id INTEGER DEFAULT NULL,
        controller_port INTEGER DEFAULT NULL,
        load_balancer_port INTEGER DEFAULT NULL,
        status TEXT,
        uptime INTEGER DEFAULT NULL,
        policy TEXT DEFAULT NULL,
        auto_restart INTEGER DEFAULT NULL,
        requested_resources BLOB DEFAULT NULL)""")
    cursor.execute("""\
        CREATE TABLE IF NOT EXISTS replicas (
        service_name TEXT,
        replica_id INTEGER,
        replica_info BLOB,
        PRIMARY KEY (service_name, replica_id))""")
    cursor.execute("""\
        CREATE TABLE IF NOT EXISTS version_specs (
        version INTEGER, 
        service_name TEXT,
        spec BLOB,
        PRIMARY KEY (service_name, version))""")
    conn.commit()


_DB = db_utils.SQLiteConn(_DB_PATH, create_table)
# Backward compatibility.
db_utils.add_column_to_table(_DB.cursor, _DB.conn, 'services',
                             'requested_resources_str', 'TEXT')
db_utils.add_column_to_table(_DB.cursor, _DB.conn, 'services',
                             'current_version',
                             f'INTEGER DEFAULT {constants.INITIAL_VERSION}')
_UNIQUE_CONSTRAINT_FAILED_ERROR_MSG = 'UNIQUE constraint failed: services.name'


# === Statuses ===
class ReplicaStatus(enum.Enum):
    """Replica status."""

    # The `sky.launch` is pending due to max number of simultaneous launches.
    PENDING = 'PENDING'

    # The replica VM is being provisioned. i.e., the `sky.launch` is still
    # running.
    PROVISIONING = 'PROVISIONING'

    # The replica VM is provisioned and the service is starting. This indicates
    # user's `setup` section or `run` section is still running, and the
    # readiness probe fails.
    STARTING = 'STARTING'

    # The replica VM is provisioned and the service is ready, i.e. the
    # readiness probe is passed.
    READY = 'READY'

    # The service was ready before, but it becomes not ready now, i.e. the
    # readiness probe fails.
    NOT_READY = 'NOT_READY'

    # The replica VM is being shut down. i.e., the `sky down` is still running.
    SHUTTING_DOWN = 'SHUTTING_DOWN'

    # The replica VM is once failed and has been deleted.
    FAILED = 'FAILED'

    # `sky.down` failed during service teardown.
    # This could mean resource leakage.
    # TODO(tian): This status should be removed in the future, at which point
    # we should guarantee no resource leakage like regular sky.
    FAILED_CLEANUP = 'FAILED_CLEANUP'

    # The replica is a spot VM and it is preempted by the cloud provider.
    PREEMPTED = 'PREEMPTED'

    # Unknown. This should never happen (used only for unexpected errors).
    UNKNOWN = 'UNKNOWN'

    @classmethod
    def failed_statuses(cls) -> List['ReplicaStatus']:
        return [cls.FAILED, cls.FAILED_CLEANUP, cls.UNKNOWN]

    @classmethod
    def launched_statuses(cls) -> List['ReplicaStatus']:
        return [cls.PENDING, cls.PROVISIONING, cls.STARTING, cls.READY]

    @classmethod
    def scale_down_decision_order(cls) -> List['ReplicaStatus']:
        # Scale down replicas in the order of replica initialization
        return [cls.PENDING, cls.PROVISIONING, cls.STARTING, cls.READY]

    def colored_str(self) -> str:
        color = _REPLICA_STATUS_TO_COLOR[self]
        return f'{color}{self.value}{colorama.Style.RESET_ALL}'


_REPLICA_STATUS_TO_COLOR = {
    ReplicaStatus.PENDING: colorama.Fore.YELLOW,
    ReplicaStatus.PROVISIONING: colorama.Fore.BLUE,
    ReplicaStatus.STARTING: colorama.Fore.CYAN,
    ReplicaStatus.READY: colorama.Fore.GREEN,
    ReplicaStatus.NOT_READY: colorama.Fore.YELLOW,
    ReplicaStatus.SHUTTING_DOWN: colorama.Fore.MAGENTA,
    ReplicaStatus.FAILED: colorama.Fore.RED,
    ReplicaStatus.FAILED_CLEANUP: colorama.Fore.RED,
    ReplicaStatus.PREEMPTED: colorama.Fore.MAGENTA,
    ReplicaStatus.UNKNOWN: colorama.Fore.RED,
}


class ServiceStatus(enum.Enum):
    """Service status as recorded in table 'services'."""

    # Controller is initializing
    CONTROLLER_INIT = 'CONTROLLER_INIT'

    # Replica is initializing and no failure
    REPLICA_INIT = 'REPLICA_INIT'

    # Controller failed to initialize / controller or load balancer process
    # status abnormal
    CONTROLLER_FAILED = 'CONTROLLER_FAILED'

    # At least one replica is ready
    READY = 'READY'

    # Service is being shutting down
    SHUTTING_DOWN = 'SHUTTING_DOWN'

    # At least one replica is failed and no replica is ready
    FAILED = 'FAILED'

    # Clean up failed
    FAILED_CLEANUP = 'FAILED_CLEANUP'

    # No replica
    NO_REPLICA = 'NO_REPLICA'

    @classmethod
    def failed_statuses(cls) -> List['ServiceStatus']:
        return [cls.CONTROLLER_FAILED, cls.FAILED_CLEANUP]

    @classmethod
    def refuse_to_terminate_statuses(cls) -> List['ServiceStatus']:
        return [cls.CONTROLLER_FAILED, cls.FAILED_CLEANUP, cls.SHUTTING_DOWN]

    def colored_str(self) -> str:
        color = _SERVICE_STATUS_TO_COLOR[self]
        return f'{color}{self.value}{colorama.Style.RESET_ALL}'

    @classmethod
    def from_replica_statuses(
            cls, replica_statuses: List[ReplicaStatus]) -> 'ServiceStatus':
        status2num = collections.Counter(replica_statuses)
        # If one replica is READY, the service is READY.
        if status2num[ReplicaStatus.READY] > 0:
            return cls.READY
        if sum(status2num[status]
               for status in ReplicaStatus.failed_statuses()) > 0:
            return cls.FAILED
        # When min_replicas = 0, there is no (provisioning) replica.
        if len(replica_statuses) == 0:
            return cls.NO_REPLICA
        return cls.REPLICA_INIT


_SERVICE_STATUS_TO_COLOR = {
    ServiceStatus.CONTROLLER_INIT: colorama.Fore.BLUE,
    ServiceStatus.REPLICA_INIT: colorama.Fore.BLUE,
    ServiceStatus.CONTROLLER_FAILED: colorama.Fore.RED,
    ServiceStatus.READY: colorama.Fore.GREEN,
    ServiceStatus.SHUTTING_DOWN: colorama.Fore.YELLOW,
    ServiceStatus.FAILED: colorama.Fore.RED,
    ServiceStatus.FAILED_CLEANUP: colorama.Fore.RED,
    ServiceStatus.NO_REPLICA: colorama.Fore.MAGENTA,
}


def add_service(name: str, controller_job_id: int, policy: str, version: int,
                requested_resources_str: str, status: ServiceStatus) -> bool:
    """Add a service in the database.

    Returns:
        True if the service is added successfully, False if the service already
        exists.
    """
    try:
        _DB.cursor.execute(
            """\
            INSERT INTO services
            (name, controller_job_id, status, policy,
            requested_resources_str, current_version)
            VALUES (?, ?, ?, ?, ?, ?)""",
            (name, controller_job_id, status.value, policy,
             requested_resources_str, version))
        _DB.conn.commit()
    except sqlite3.IntegrityError as e:
        if str(e) != _UNIQUE_CONSTRAINT_FAILED_ERROR_MSG:
            raise RuntimeError('Unexpected database error') from e
        return False
    return True


def remove_service(service_name: str) -> None:
    """Removes a service from the database."""
    _DB.cursor.execute("""\
        DELETE FROM services WHERE name=(?)""", (service_name,))
    _DB.conn.commit()


def set_service_uptime(service_name: str, uptime: int) -> None:
    """Sets the uptime of a service."""
    _DB.cursor.execute(
        """\
        UPDATE services SET
        uptime=(?) WHERE name=(?)""", (uptime, service_name))
    _DB.conn.commit()


def set_service_status(service_name: str, status: ServiceStatus) -> None:
    """Sets the service status."""
    _DB.cursor.execute(
        """\
        UPDATE services SET
        status=(?) WHERE name=(?)""", (status.value, service_name))
    _DB.conn.commit()


def set_service_controller_port(service_name: str,
                                controller_port: int) -> None:
    """Sets the controller port of a service."""
    _DB.cursor.execute(
        """\
        UPDATE services SET
        controller_port=(?) WHERE name=(?)""", (controller_port, service_name))
    _DB.conn.commit()


def set_service_load_balancer_port(service_name: str,
                                   load_balancer_port: int) -> None:
    """Sets the load balancer port of a service."""
    _DB.cursor.execute(
        """\
        UPDATE services SET
        load_balancer_port=(?) WHERE name=(?)""",
        (load_balancer_port, service_name))
    _DB.conn.commit()


def _get_service_from_row(row) -> Dict[str, Any]:
    (name, controller_job_id, controller_port, load_balancer_port, status,
     uptime, policy, _, requested_resources, requested_resources_str,
     current_version) = row[:11]
    return {
        'name': name,
        'controller_job_id': controller_job_id,
        'controller_port': controller_port,
        'load_balancer_port': load_balancer_port,
        'status': ServiceStatus[status],
        'uptime': uptime,
        'policy': policy,
        'version': current_version,
        # TODO(tian): Backward compatibility.
        # Remove after 2 minor release, 0.6.0.
        'requested_resources': pickle.loads(requested_resources)
                               if requested_resources is not None else None,
        'requested_resources_str': requested_resources_str,
    }


def get_services() -> List[Dict[str, Any]]:
    """Get all existing service records."""
    rows = _DB.cursor.execute('SELECT * FROM services').fetchall()
    records = []
    for row in rows:
        records.append(_get_service_from_row(row))
    return records


def get_service_from_name(service_name: str) -> Optional[Dict[str, Any]]:
    """Get all existing service records."""
    rows = _DB.cursor.execute('SELECT * FROM services WHERE name=(?)',
                              (service_name,)).fetchall()
    for row in rows:
        return _get_service_from_row(row)
    return None


def get_service_version(service_name: str) -> Optional[int]:
    """Get the version of a service."""
    service = get_service_from_name(service_name)
    if service is None:
        return None
    return service['version']


def get_glob_service_names(
        service_names: Optional[List[str]] = None) -> List[str]:
    """Get service names matching the glob patterns.

    Args:
        service_names: A list of glob patterns. If None, return all service
            names.

    Returns:
        A list of non-duplicated service names.
    """
    if service_names is None:
        rows = _DB.cursor.execute('SELECT name FROM services').fetchall()
    else:
        rows = []
        for service_name in service_names:
            rows.extend(
                _DB.cursor.execute(
                    'SELECT name FROM services WHERE name GLOB (?)',
                    (service_name,)).fetchall())
    return list({row[0] for row in rows})


# === Replica functions ===
def add_or_update_replica(service_name: str, replica_id: int,
                          replica_info: 'replica_managers.ReplicaInfo') -> None:
    """Adds a replica to the database."""
    _DB.cursor.execute(
        """\
        INSERT OR REPLACE INTO replicas
        (service_name, replica_id, replica_info)
        VALUES (?, ?, ?)""",
        (service_name, replica_id, pickle.dumps(replica_info)))
    _DB.conn.commit()


def remove_replica(service_name: str, replica_id: int) -> None:
    """Removes a replica from the database."""
    _DB.cursor.execute(
        """\
        DELETE FROM replicas
        WHERE service_name=(?)
        AND replica_id=(?)""", (service_name, replica_id))
    _DB.conn.commit()


def get_replica_info_from_id(
        service_name: str,
        replica_id: int) -> Optional['replica_managers.ReplicaInfo']:
    """Gets a replica info from the database."""
    rows = _DB.cursor.execute(
        """\
        SELECT replica_info FROM replicas
        WHERE service_name=(?)
        AND replica_id=(?)""", (service_name, replica_id)).fetchall()
    for row in rows:
        return pickle.loads(row[0])
    return None


def get_replica_infos(
        service_name: str) -> List['replica_managers.ReplicaInfo']:
    """Gets all replica infos of a service."""
    rows = _DB.cursor.execute(
        """\
        SELECT replica_info FROM replicas
        WHERE service_name=(?)""", (service_name,)).fetchall()
    return [pickle.loads(row[0]) for row in rows]


def total_number_provisioning_replicas() -> int:
    """Returns the total number of provisioning replicas."""
    rows = _DB.cursor.execute('SELECT replica_info FROM replicas').fetchall()
    provisioning_count = 0
    for row in rows:
        replica_info: 'replica_managers.ReplicaInfo' = pickle.loads(row[0])
        if replica_info.status == ReplicaStatus.PROVISIONING:
            provisioning_count += 1
    return provisioning_count


# === Version functions ===
def add_version(service_name: str) -> int:
    """Adds a version to the database."""

    _DB.cursor.execute(
        """\
        INSERT INTO version_specs
        (version, service_name, spec)
        VALUES (
            (SELECT COALESCE(MAX(version), 0) + 1 FROM
            version_specs WHERE service_name = ?), ?, ?)
        RETURNING version""", (service_name, service_name, pickle.dumps(None)))

    inserted_version = _DB.cursor.fetchone()[0]
    _DB.conn.commit()

    return inserted_version


def add_or_update_version(service_name: str, version: int,
                          spec: 'service_spec.SkyServiceSpec') -> None:
    _DB.cursor.execute(
        """\
        INSERT or REPLACE INTO version_specs
        (service_name, version, spec)
        VALUES (?, ?, ?)""", (service_name, version, pickle.dumps(spec)))
    _DB.cursor.execute(
        """\
        UPDATE services SET
        current_version=(?) WHERE name=(?)""", (version, service_name))
    _DB.conn.commit()


def remove_service_versions(service_name: str) -> None:
    """Removes a replica from the database."""
    _DB.cursor.execute(
        """\
        DELETE FROM version_specs
        WHERE service_name=(?)""", (service_name,))
    _DB.conn.commit()


def get_spec(service_name: str,
             version: int) -> Optional['service_spec.SkyServiceSpec']:
    """Gets spec from the database."""
    rows = _DB.cursor.execute(
        """\
        SELECT spec FROM version_specs
        WHERE service_name=(?)
        AND version=(?)""", (service_name, version)).fetchall()
    for row in rows:
        return pickle.loads(row[0])
    return None


def delete_version(service_name: str, version: int) -> None:
    """Deletes a version from the database."""
    _DB.cursor.execute(
        """\
        DELETE FROM version_specs
        WHERE service_name=(?)
        AND version=(?)""", (service_name, version))
    _DB.conn.commit()



"""Statuses enum for SkyPilot resources."""

import enum

import colorama


class ClusterStatus(enum.Enum):
    """Cluster status as recorded in table 'clusters'."""
    # NOTE: these statuses are as recorded in our local cache, the table
    # 'clusters'.  The actual cluster state may be different (e.g., an UP
    # cluster getting killed manually by the user or the cloud provider).

    # Initializing.  This means a backend.provision() call has started but has
    # not successfully finished. The cluster may be undergoing setup, may have
    # failed setup, may be live or down.
    INIT = 'INIT'

    # The cluster is recorded as up.  This means a backend.provision() has
    # previously succeeded.
    UP = 'UP'

    # Stopped.  This means a `sky stop` call has previously succeeded.
    STOPPED = 'STOPPED'

    def colored_str(self):
        color = _STATUS_TO_COLOR[self]
        return f'{color}{self.value}{colorama.Style.RESET_ALL}'


_STATUS_TO_COLOR = {
    ClusterStatus.INIT: colorama.Fore.BLUE,
    ClusterStatus.UP: colorama.Fore.GREEN,
    ClusterStatus.STOPPED: colorama.Fore.YELLOW,
}


class StorageStatus(enum.Enum):
    """Storage status as recorded in table 'storage'."""

    # Initializing and uploading storage
    INIT = 'INIT'

    # Initialization failed
    INIT_FAILED = 'INIT_FAILED'

    # Failed to Upload to Cloud
    UPLOAD_FAILED = 'UPLOAD_FAILED'

    # Finished uploading, in terminal state
    READY = 'READY'
